;;; Wbuild mode commands for Emacs
;;;
;;; Wbuild mode is a combination of latex-mode and C mode,
;;; where some commands depend on the context.
;;; Many keys are bound to functions that first check whether the
;;; point is in a code "chunk" or a documentation "chunk".
;;;
;;; Author: Bert Bos <bert@let.rug.nl>
;;; Date: 31 Oct 1994
;;;
;;; 951127: Changed fontification patterns, removed 
;;;	my-fill-region-as paragraph [BB]

(require 'tex-mode)
(if (not (string-match "XEmacs\\|Lucid" emacs-version))
    (require 'c-mode)
  (require 'cc-mode)
  (require 'font-lock))
;;(load-file "/users1/bert/Emacs/cweb.elc")
;;(load-file "~/Emacs/myfill.elc")

(provide 'wbuild-mode)

;;;
;;; Some regular expressions
;;;

(defconst wbuild-code-re
  "^[ \t]*@\\(proc\\|def\\|var\\|type\\|incl\\|trans\\)[ \t]"
  "Regular expression matching the start of a code chunk.")

(defconst wbuild-text-re "^[ \t]*@[ \t]"
  "Regular expression matching the start of a text chunk.")

(defconst wbuild-chunk-re (concat "^[ \t]*@")
  "Regular expression matching the start of a chunk.")

(defconst wbuild-paragraph-start-re
  (concat wbuild-text-re "\\|" wbuild-code-re "\\|" 
	  "^[ \t]*$\\|^[\f\\\\%]\\|^[ \t]*\\$\\$")
  "Regular expression matching the start of a paragraph")

(defconst wbuild-paragraph-sep-re "^[ \t]*$"
  "Regular expression matching the start of a paragraph")

(defconst wbuild-comment-start-skip
  "\\([^\\]\\)\\(%+[ \t]*\\|/\\*+[ \t]*\\)"
;;;  "\\(\\(^\\|[^\\]\\)\\(\\\\\\\\\\)*\\)\\(%+ *\\)\\|/\\*+ *"
  "Regular expression matching the start of a comment.")

(defvar wbuild-mode-hook nil
  "*Hook called by `wbuild-mode'.")

;;;
;;; Default key bindings (combine tex-mode-map, c-mode-map and
;;; extra's) and syntax entries.
;;;

(defvar wbuild-mode-map nil "Keymap used in wbuild-mode")
(defvar wbuild-mode-syntax-table nil "Syntax table used in wbuild-mode")

(if wbuild-mode-map
    ()
  ;;(setq wbuild-mode-map (copy-keymap tex-mode-map))
  (setq wbuild-mode-map (copy-keymap c-mode-map))
  (define-key wbuild-mode-map "{"      'wbuild-electric-c-brace)
  (define-key wbuild-mode-map "}"      'wbuild-electric-c-brace)
  (define-key wbuild-mode-map ";"      'wbuild-electric-c-semi)
  (define-key wbuild-mode-map ":"      'wbuild-electric-c-terminator)
  (define-key wbuild-mode-map "\e\C-h" 'mark-c-function)
  (define-key wbuild-mode-map "\e\C-q" 'indent-c-exp)
  (define-key wbuild-mode-map "\177"   'delete-char)
  ;;
  (define-key wbuild-mode-map "\C-\\"  'pop-pending-list)
  (define-key wbuild-mode-map "\M-\\"  'into-pending-list)
  (define-key wbuild-mode-map "\M-y"   'new-yank-pop)
  ;;(define-key wbuild-mode-map "\M-w"   'indicate-and-copy-region)
  (define-key wbuild-mode-map "\M-\""  'ditto)
  (define-key wbuild-mode-map "\M-n"   'wbuild-forward-module)
  (define-key wbuild-mode-map "\M-p"   'wbuild-backward-module)
  (define-key wbuild-mode-map "\M-q"   'wbuild-fill-paragraph)
  (define-key wbuild-mode-map "\r"     'wbuild-newline)
  (define-key wbuild-mode-map "\t"     'wbuild-indent)
  (define-key wbuild-mode-map "\M-;"   'wbuild-indent-for-comment)
  (define-key wbuild-mode-map "\""     'self-insert-command))

(if wbuild-mode-syntax-table
    ()
  (setq wbuild-mode-syntax-table (copy-syntax-table c-mode-syntax-table))
  (modify-syntax-entry ?\' "." wbuild-mode-syntax-table)
  )

;;;
;;; Definition of wbuild-mode (major mode)
;;;

(defvar wbuild-mode nil)

(defun wbuild-mode ()
  "Major mode wbuild-mode is a combination of c-mode and TeX-mode, with
some extras. Some commands only work in documentation chunks, others
only in code chunk. The extras include \\[forward-module] and
\\[backward-module] for relative module movement. The automatic \"
feature is disabled."
  (interactive)
  (kill-all-local-variables)
  (c-mode)
  (use-local-map wbuild-mode-map)
  ;;(setq local-abbrev-table wbuild-mode-abbrev-table)
  (setq local-abbrev-table c-mode-abbrev-table)
  (set-syntax-table wbuild-mode-syntax-table)
  (setq mode-name "wbuild")
  (setq major-mode 'wbuild-mode)
  ;;
  (make-local-variable 'paragraph-start)
  (setq paragraph-start wbuild-paragraph-start-re)
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate wbuild-paragraph-sep-re)
  ;;(make-local-variable 'comment-start)
  ;;(setq comment-start "/* ")
  ;;(make-local-variable 'comment-end)
  ;;(setq comment-end " */")
  ;;(make-local-variable 'comment-column)
  ;;(setq comment-column 32)
  ;;(make-local-variable 'comment-start-skip)
  ;;(setq comment-start-skip wbuild-comment-start-skip)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'wbuild-comment-indent)
  ;;(make-local-variable 'parse-sexp-ignore-comments)
  ;;(setq parse-sexp-ignore-comments t)
  ;;
  ;;(make-local-variable 'paragraph-ignore-fill-prefix)
  ;;(setq paragraph-ignore-fill-prefix t)
  ;;(make-local-variable 'indent-line-function)
  ;;(setq indent-line-function 'c-indent-line)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  ;;
  ;;
  (run-hooks 'text-mode-hook 'tex-mode-hook 'latex-mode-hook
	     'wbuild-mode-hook))

;;;
;;; New commands: jump to next/previous text chunks
;;;

(defun wbuild-forward-module (arg)
  "Advance past next wbuild module beginning; with ARG, repeat ARG times."
  (interactive "p")
  (wbuild-move-to-module arg))

(defun wbuild-backward-module (arg)
  "Advance to previous wbuild module beginning; with ARG, repeat ARG times."
  (interactive "p")
  (wbuild-move-to-module (- arg)))

(defun wbuild-move-to-module (arg)
  "Move to ARG'th module forward if > 0, backward if < 0."
  (while (> arg 0)
    (re-search-forward wbuild-text-re)
    (setq arg (1- arg)))
  (while (< arg 0)
    (re-search-backward wbuild-text-re)
    (setq arg (1+ arg))))

;;;
;;; New command: visually indicate a region that is copied
;;;

;;; The next lines of code are controversial --- they seem to be the
;;; best way to do a short wait and redraw the screen with standard
;;; emacs primitives --- but the short wait is a "busy wait". On a
;;; faster machine, it would be better to install the function
;;; sit-for-millisecs found in sunfns.c (if not already installed) and
;;; to say (sit-for-millisecs 100) instead. On a slower machine, do
;;; the call-process only once. On a still slower machine, (sit-for 1)
;;; is probably best.
;;;
;;; (call-process "echo" nil nil t)
;;; (call-process "echo" nil nil t))

(defun indicate-region ()
  "Bounce cursor to mark and back again."
  (let ((point-save (point)))
    (unwind-protect
        (progn
	  (goto-char (mark))
	  (sit-for 1))
      (goto-char point-save))))

(defun indicate-and-copy-region (beg end)
  "Indicate current region, then copy it to the kill ring."
  (interactive "r")(indicate-region)(copy-region-as-kill beg end))

;;;
;;; New commands: In text mode, check for balanced braces and quotes.
;;;

(defun wbuild-newline (arg)
  "If previous character is newline and no ARG, check for unbalanced braces
and/or dollar signs in previous paragraph. If ARG is \\[universal-argument],
do a single newline; otherwise do ordinary newline."
  (interactive "*P")
  (cond
   ((and (eq (preceding-char) ?\n) (not arg) (wbuild-text-chunk-p))
    (if (not (save-excursion
	       (tex-validate-region
		(save-excursion (search-backward "\n\n" nil 'move) (point))
		(point))))
	(message "Paragraph being closed appears to contain a mismatch"))
    (newline arg))
   ((listp arg)
    (newline))
   (t
    (newline arg))))

(defun wbuild-text-chunk-p ()
  "Return non-nil if point is in a text chunk."
  (save-excursion
    (end-of-line)
    (re-search-backward wbuild-chunk-re 0 t)
    (looking-at "@[ \t\n]")))

;;;
;;; Following function currently not used
;;;

(defun wbuild-font-lock-after-change-function (beg end old-len)
  "Called when any modification is made to buffer text."
  (if (not (wbuild-text-chunk-p))
      (font-lock-after-change-function beg end old-len)))

;;;
;;; The tab key is usually bound to wbuild-indent. It uses
;;; indent-relative or c-indent-command, depending on context.
;;;

(defun wbuild-indent (&optional arg)
  "Depending on context, either inserts a tab or indents the current
line according to the level of braces. The former happens in text
chunks, the latter in C code."
  (interactive "P")
  (if (or (save-excursion (beginning-of-line) (looking-at wbuild-chunk-re))
	  (wbuild-text-chunk-p))
      (indent-relative)		; Indent text line or line with keyword
    (save-restriction
      (widen)
      (let
	  ((matchend
	    (save-excursion
	     (if (re-search-backward wbuild-chunk-re 0 t)
		 (match-end 0)
	       0)))
	   (end
	    (save-excursion
	      (if (re-search-forward wbuild-chunk-re (point-max) t)
		  (match-beginning 0)
		(point-max)))))
	(narrow-to-region matchend end)
	(c-indent-line arg)))))

;;;
;;; New command: wbuild-fill-paragraph either fills text or indents C code
;;;

(defun wbuild-fill-paragraph (arg)
  "In text chunks, fills text in a paragraph. In code chunks, does nothing."
  (interactive "P")
  (if (wbuild-text-chunk-p)
      (wbuild-fill arg)))

(defun wbuild-fill (arg)
  (let ((fill-prefix "")
	(beg 0))
    (forward-paragraph)			; just in case we're on the 1st line
    (backward-paragraph)		; go to start of paragraph
    (skip-chars-forward "\n")		; forward if on empty line
    (setq beg (point))			; store start of paragraph
    (beginning-of-line)
    (delete-horizontal-space)
    (insert fill-prefix)
    (forward-paragraph)			; find end of paragraph
    (or (bolp) (newline 1))		; add a newline, if needed
    ;;(my-fill-region-as-paragraph beg (point) arg)
    (fill-region-as-paragraph beg (point) arg)
))

;;;
;;; New commands: make {. }, : and : act specially in code chunks
;;;

(defun wbuild-electric-c-brace (arg)
  "In code chunks, insert character and correct line's indentation."
  (interactive "P")
  (if (wbuild-text-chunk-p)
      (insert last-command-char)
    (electric-c-brace arg)))

(defun wbuild-electric-c-semi (arg)
  "In code chunks, insert character and correct line's indentation."
  (interactive "P")
  (if (wbuild-text-chunk-p)
      (insert last-command-char)
    (electric-c-semi arg)))

(defun wbuild-electric-c-terminator (arg)
  "In code chunks, insert character and correct line's indentation."
  (interactive "P")
  (if (wbuild-text-chunk-p)
      (insert last-command-char)
    (c-electric-colon arg)))

;;;
;;; Hook for indenting comment lines.
;;;

(defun wbuild-comment-indent ()
  "Indent either the current TeX comment or C comment."
  (if (wbuild-text-chunk-p)
      (tex-comment-indent)
    (c-comment-indent)))

(defun wbuild-indent-for-comment ()
  "Indent this line's comment to comment column, or insert an empty comment.
In text chunks, the comment is a TeX comment (%...), in code chunks it is a
C comment (/*...*/)"
  (interactive "*")
  (if (wbuild-text-chunk-p)
      (let ((comment-start "% ")
	    (comment-end "")
	    (comment-column 0))
	(indent-for-comment))
    (indent-for-comment)))

;;;
;;; font-lock patterns
;;;

(require 'font-lock)

(make-face 'font-lock-chunk-face)
(or (face-differs-from-default-p 'font-lock-chunk-face)
    (progn
      (copy-face 'bold 'font-lock-chunk-face)
      ;;(set-face-background 'font-lock-chunk-face "blue3")
      ;;(set-face-background 'font-lock-chunk-face "pink")
      (set-face-background 'font-lock-chunk-face "yellow")
      (set-face-foreground 'font-lock-chunk-face "#aa2400")
      ))

(put wbuild-mode 'font-lock-keywords 'wbuild-font-lock-keywords)

(defconst wbuild-font-lock-keywords-1 nil
 "For consideration as a value of `wbuild-font-lock-keywords'.
This does fairly subdued highlighting.")

(defconst wbuild-font-lock-keywords-2 nil
 "For consideration as a value of `wbuild-font-lock-keywords'.
This does a lot more highlighting.")

(let ((storage "auto\\|extern\\|register\\|static\\|volatile")
      (prefixes "unsigned\\|short\\|long\\|const")
      (types (concat "int\\|long\\|char\\|float\\|double\\|void\\|struct\\|"
		     "union\\|enum\\|typedef"))
      (ctoken "\\(\\sw\\|\\s_\\|[:~*&]\\)+")
      (wkeywords (concat "@[Pp][Uu][Bb][Ll][Ii][Cc]\\|"
			 "@[Pp][Rr][Ii][Vv][Aa][Tt][Ee]-[Cc][Oo][Nn][Ss][Tt][Rr][Aa][Ii][Nn][Tt][Ss]\\|"
			 "@[Pp][Rr][Ii][Vv][Aa][Tt][Ee]\\|"
			 "@[Mm][Ee][Tt][Hh][Oo][Dd][Ss]\\|"
			 "@[Ii][Mm][Pp][Oo][Rr][Tt][Ss]\\|"
			 "@[Ee][Xx][Pp][Oo][Rr][Tt][Ss]\\|"
			 "@[Cc][Ll][Aa][Ss][Ss]\\|"
			 "@[Cc][Oo][Nn][Ss][Tt][Rr][Aa][Ii][Nn][Tt][Ss]\\|"
			 "@[Cc][Ll][Aa][Ss][Ss][Vv][Aa][Rr][Ss]\\|"
			 "@[Uu][Tt][Ii][Ll][Ii][Tt][Ii][Ee][Ss]\\|"
			 "@[Tt][Rr][Aa][Nn][Ss][Ll][Aa][Tt][Ii][Oo][Nn][Ss]\\|"
			 "@[Aa][Cc][Tt][Ii][Oo][Nn][Ss]"))
      )
  (setq wbuild-font-lock-keywords-1 (purecopy
   (list
    ;;
    '("^[ \t]*@[A-Za-z]+" 0 font-lock-keyword-face)
    ;;
    ;; fontify chunk names as chunk names
    (list (concat "^[ \t]*\\(" wkeywords "\\)\\b") 0 'font-lock-chunk-face t)
    ;;
    ;; fontify code fragments in text chunks
    ;; '("|[^|\n]+|" 0 font-lock-function-name-face t)
    ;;
    ;; fontify text chunks
    ;; '("^\\([ \t]*@[ \t][^@]*[^@ \t\n]\\)[ \t\n]*" 1 font-lock-doc-string-face t)
    '("^\\([ \t]*@[ \t]\\([^@]\\|@@\\)*\\)@" 1 font-lock-doc-string-face t)
    ;; fontify names being defined.
    (list (concat "^#[ \t]*\\(define\\|undef\\)[ \t]+" ctoken)
	  2 'font-lock-function-name-face)
    (list (concat "^[ \t]*@\\([Dd][Ee][Ff]\\|[Tt][Yy][Pp][Ee]\\)[ \t]+" ctoken)
	  2 'font-lock-function-name-face)
    ;;
    ;; fontify other preprocessor lines.
    '("^#[ \t]*\\(if\\|ifn?def\\|elif\\)[ \t]+\\([^\n]+\\)"
      2 font-lock-function-name-face t)
    ;;
    ;; fontify the filename in #include <...>
    ;; don't need to do this for #include "..." because those were
    ;; already fontified as strings by the syntactic pass.
    ;; (Changed to not include the <> in the face, since "" aren't.)
    ;; '("^#[ \t]*include[ \t]+<\\([^>\"\n]+\\)>" 1 font-lock-string-face)
    '("^[ \t]*@[Ii][Nn][Cc][Ll][ \t]+<\\([^>\"\n]+\\)>"
      1 font-lock-string-face)
    ;;
    ;; fontify preprocessor directives as comments.
    '("^#.*" . font-lock-comment-face)
    ;;
    ;; fontify the names of functions being defined.
    ;; I think this should be fast because it's anchored at bol, but it's not.
    ;;(list (concat
    ;;	   "^\\(" ctoken "[ \t]+\\)?"	; type specs; there can be no
    ;;	   "\\(" ctoken "[ \t]+\\)?"	; more than 3 tokens, right?
    ;;	   "\\(" ctoken "[ \t]+\\)?"
    ;;	   "\\([*&]+[ \t]*\\)?"		; pointer
    ;;	   "\\(" ctoken "\\)[ \t]*(")	; name
    ;;	  8 'font-lock-function-name-face)
    ;;
    ;; This is faster but not by much.  I don't see why not.
    ;; (list (concat "^\\(" ctoken "\\)[ \t]*(")
    ;;       1 'font-lock-function-name-face)
    ;;
    ;; fontify function names being defined
    (list (concat "^[ \t]*@proc[ \t]+\\(" ctoken "[ \t]+\\)*\\("
		  ctoken "\\)[ \t]*(")
	  3 'font-lock-function-name-face)
    (list (concat "^[ \t]*@proc[ \t]+\\(" ctoken "[ \t]+\\)*\\("
		  ctoken "\\)[ \t]*$")
	  3 'font-lock-function-name-face)
    ;;
    ;; fontify variable names being defined
    ;(list (concat "^[ \t]*@var[ \t]+\\(" ctoken "[ \t]+\\)*" ctoken "[ \t]*=")
    ;;	  3 'font-lock-function-name-face)
    ;(list (concat "^[ \t]*@var[ \t]+\\(" ctoken "[ \t]+\\)*" ctoken "[ \t]*$")
    ;;	  3 'font-lock-function-name-face)
    ;;
    ;; Fontify structure names (in structure definition form).
    (list (concat "^\\(typedef[ \t]+struct\\|struct\\|static[ \t]+struct\\)"
		  "[ \t]+\\(" ctoken "\\)[ \t]*\\(\{\\|$\\)")
	  2 'font-lock-function-name-face)
    ;;
    ;; fontify comments
    ;;'("/\\*+.*\\*+/" 0 font-lock-comment-face t)
    '("/\\*\\([^*]\\|\\*[^/]\\)*\\*/" 0 font-lock-comment-face)
    ;;
    ;; Fontify case clauses.  This is fast because its anchored on the left.
    '("case[ \t]+\\(\\(\\sw\\|\\s_\\)+\\):". 1)
    '("\\<\\(default\\):". 1)
    ;;
    ;; fontify strings
    ;;'("'\\(.\\|..\\|...\\)'" 1 font-lock-string-face t)
    '("'\\(.\\|..\\|...\\)'" 0 font-lock-string-face)
    ;;'("\"\\([^\"\n]\\|\\\\\"\\|\\\\\n\\)*\"" 0 font-lock-string-face t)
    '("\"\\([^\"\n]\\|\\\\\"\\|\\\\\n\\)*\"" 0 font-lock-string-face)
    )))

  (setq wbuild-font-lock-keywords-2 (purecopy
   (append wbuild-font-lock-keywords-1
    (list
     ;;
     ;; fontify all storage classes and type specifiers
     (cons (concat "\\<\\(" storage "\\)\\>") 'font-lock-type-face)
     (cons (concat "\\<\\(" types "\\)\\>") 'font-lock-type-face)
     (cons (concat "\\<\\(" prefixes "[ \t]+" types "\\)\\>")
	   'font-lock-type-face)
     ;;
     ;; fontify all builtin tokens
     (cons (concat
	    "[ \t]\\("
	    (mapconcat 'identity
	     '("for" "while" "do" "return" "goto" "case" "break" "switch"
	       "if" "then" "else if" "else" "return" "continue" "default"
	       )
	     "\\|")
	    "\\)[ \t\n(){};,]")
	   1)
     ;;
     ;; fontify case targets and goto-tags.  This is slow because the
     ;; expression is anchored on the right.
     "\\(\\(\\sw\\|\\s_\\)+\\):"
     ;;
     ;; Fontify variables declared with structures, or typedef names.
     '("}[ \t*]*\\(\\(\\sw\\|\\s_\\)+\\)[ \t]*[,;]"
       1 font-lock-function-name-face)
     ;;
     ;; Fontify global variables without a type.
;     '("^\\([_a-zA-Z0-9:~*]+\\)[ \t]*[[;={]" 1 font-lock-function-name-face)

     ))))
  )


;; default to the gaudier variety
(defconst wbuild-font-lock-keywords wbuild-font-lock-keywords-2
  "Additional expressions to highlight in wbuild mode.")
;(defconst wbuild-font-lock-keywords wbuild-font-lock-keywords-1
;  "Additional expressions to highlight in wbuild mode.")

